"""Combobox that displays a list of items to be checked."""

from __future__ import annotations

import contextlib
import logging
from typing import Generic, TypeVar, cast

from PySide6 import QtGui
from PySide6.QtCore import (
    QEvent,
    QModelIndex,
    QObject,
    QPoint,
    Qt,
    Signal,
    Slot,
)
from PySide6.QtGui import (
    QFontMetrics,
    QIcon,
    QMouseEvent,
    QPainter,
    QResizeEvent,
    QStandardItem,
    QStandardItemModel,
)
from PySide6.QtWidgets import (
    QComboBox,
    QListView,
    QStyle,
    QStyleOptionComboBox,
    QWidget,
)

from qute_style.style import get_color
from qute_style.widgets.custom_icon_engine import CustomIconEngine, PixmapStore

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name

ItemData = TypeVar("ItemData")


class TooManyItemsError(Exception):
    """Raised when more than one item shall be checked in single mode."""


class StyledComboBox(QComboBox):
    """Combobox with custom arrow drawing."""

    def __init__(self, parent: QWidget | None = None) -> None:
        """Create a new StyledComboBox."""
        super().__init__(parent)
        # options below are needed, otherwise the rounded
        # corners of dropdowns cannot be drawn correctly
        self.view().window().setWindowFlags(
            Qt.WindowType.Popup | Qt.WindowType.FramelessWindowHint
        )
        self.view().window().setAttribute(
            Qt.WidgetAttribute.WA_TranslucentBackground
        )
        # stop combobox contents from limiting width of whole application
        self.setSizeAdjustPolicy(
            QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon
        )

    def paintEvent(  # pylint: disable=invalid-name
        self, event: QtGui.QPaintEvent
    ) -> None:
        """Overwrite method to draw custom arrow."""
        # Draw ComboBox like defined in style sheet
        super().paintEvent(event)

        # Paint custom arrow above ComboBox
        painter = QPainter()
        painter.begin(self)

        # Set correct color
        if self.isEnabled():
            color = get_color("foreground")
        else:
            color = get_color("fg_disabled")

        # Define size of arrow depending on ComboBox SubControl
        opt = QStyleOptionComboBox()
        rect = self.style().subControlRect(
            QStyle.ComplexControl.CC_ComboBox,
            opt,
            QStyle.SubControl.SC_ComboBoxArrow,
            None,
        )
        radius = int(rect.width() * 0.5)

        # Get correct scale and pixmap
        scale = int(painter.device().devicePixelRatio())
        pixmap = PixmapStore.inst().get_pixmap(
            ":/svg_icons/expand_more.svg",
            radius * scale,
            radius * scale,
            color,
        )
        # Draw arrow above existing ComboBox
        painter.setCompositionMode(
            QPainter.CompositionMode.CompositionMode_SourceOver
        )
        # Draw arrow in rect 25x25 at the end, with size: radius x radius
        # x-axis: in the middle of arrow rect with spacing 2
        # y-axis: in the middle of arrow rect
        painter.drawPixmap(
            int(self.width() - 0.5 * (4 + rect.width() + radius)),
            int(0.5 * (self.height() - radius)),
            radius,
            radius,
            pixmap,
        )
        painter.end()


class CheckableComboBox(StyledComboBox, Generic[ItemData]):
    """Combobox that displays a list of items to be checked."""

    dataChanged = Signal(dict, name="dataChanged")

    def __init__(self, parent: QWidget | None = None) -> None:
        """Create a new CheckableComboBox."""
        super().__init__(parent)

        # default text that is shown when nothing is selected
        self._default_text = self.tr("Keine Auswahl")

        # sets the single mode so that only one class can be selected.
        self._single = False

        # Make the combo editable to set a custom text, but readonly
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)

        # Update the text when an item is toggled
        self.model().dataChanged.connect(self.handle_data_change)

        # Prevent popup from closing when clicking on an item
        self.view().viewport().installEventFilter(self)

        # Hide and show popup when clicking the line edit
        self.lineEdit().installEventFilter(self)
        self.popup_open = False

        # add some spacing between the items
        cast(QListView, self.view()).setSpacing(2)

    @property
    def single_mode(self) -> bool:
        """Return if only one item can be checked at a time (single mode)."""
        return self._single

    @single_mode.setter
    def single_mode(self, single_mode: bool) -> None:
        """Set the single mode."""
        log.debug("Setting single mode to %s", single_mode)
        self._single = single_mode

    def resizeEvent(  # pylint: disable=invalid-name
        self, event: QResizeEvent
    ) -> None:
        """Recompute text when CheckableCombobox is resized."""
        super().resizeEvent(event)
        self.update_text()

    def eventFilter(  # pylint: disable=invalid-name
        self, obj: QObject, event: QEvent
    ) -> bool:
        """Filter events to show popup and set check states."""
        if event.type() == QEvent.Type.MouseButtonRelease:
            event = cast(QMouseEvent, event)
            if obj is self.lineEdit():
                log.debug(
                    "User clicked on line edit, popup is open: %s",
                    self.popup_open,
                )
                if self.popup_open:
                    # popup is closed automatically on click on Combobox
                    # we just need to set the correct state
                    self.popup_open = False
                else:
                    self.showPopup()
                    self.popup_open = True
                return True

            if obj is self.view().viewport():
                with contextlib.suppress(AttributeError):
                    self._check_item_at_pos(event.pos())
                    return True
        return False

    def hidePopup(self) -> None:  # pylint: disable=invalid-name
        """Set the state correctly when the popup is hidden from outside."""
        log.debug("Hiding popup")
        self.popup_open = False
        super().hidePopup()

    def _check_item_at_pos(self, pos: QPoint) -> None:
        """Toggle the CheckState at the given pos."""
        index = self.view().indexAt(pos)
        item = cast(QStandardItemModel, self.model()).item(index.row())
        if item.checkState() == Qt.CheckState.Checked:
            item.setCheckState(Qt.CheckState.Unchecked)
        else:
            item.setCheckState(Qt.CheckState.Checked)

    @Slot(
        QModelIndex,
        QModelIndex,
        "QVector<int>",  # type: ignore
        name="handle_data_change",
    )
    def handle_data_change(
        self,
        start: QModelIndex,
        end: QModelIndex,
        roles: list[Qt.ItemDataRole],
    ) -> None:
        """Handle a data change event to handle single_mode."""
        if (
            self._single
            and Qt.ItemDataRole.CheckStateRole in roles
            and Qt.CheckState(start.data(Qt.ItemDataRole.CheckStateRole))
            == Qt.CheckState.Checked
        ):
            log.debug("Unchecking other items, since single mode is active.")
            # we should never edit check state for two indexes at the same time
            assert start == end
            for idx in range(self.model().rowCount()):
                if idx != start.row():
                    index = self.model().index(idx, 0, QModelIndex())
                    self.model().setData(
                        index,
                        Qt.CheckState.Unchecked,
                        Qt.ItemDataRole.CheckStateRole,
                    )
        self.update_text()
        self.send_current_state()

    def update_text(self) -> None:
        """Update the texts."""
        text = self._get_text()

        # Compute elided text (with "...")
        metrics = QFontMetrics(self.lineEdit().font())
        elided_text = metrics.elidedText(
            text, Qt.TextElideMode.ElideRight, self.lineEdit().width()
        )
        self.lineEdit().setText(elided_text)

    def send_current_state(self) -> None:
        """Emit the current state via dataChanged."""
        current_state: dict[str | int, Qt.CheckState] = {
            self.model().item(i).data(): (self.model().item(i).checkState())
            for i in range(self.model().rowCount())
        }
        self.dataChanged.emit(current_state)

    def _get_text(self) -> str:
        """Return the text that is shown at the top of the combobox."""
        texts = [
            str(cast(QStandardItemModel, self.model()).item(idx).data())
            for idx in range(self.model().rowCount())
            if cast(QStandardItemModel, self.model()).item(idx).checkState()
            == Qt.CheckState.Checked
        ]

        text = ", ".join(texts)
        # set to no index otherwise the state icon is displayed from
        # the first item in the combo box text which is selected by default
        self.setCurrentIndex(-1)
        if not text:
            text = self._default_text
        return text

    def addItem(  # type: ignore # pylint: disable=invalid-name
        self,
        text: str,
        data: ItemData | None = None,
        icon_path: str | None = None,
        icon_color: str | None = None,
    ) -> None:
        """Add an Item to the Combobox."""
        if icon_color and not icon_path:
            raise AssertionError(
                "Color can only be passed together with an icon (path)"
            )
        item = QStandardItem(text)
        item.setData(data)
        item.setFlags(
            Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable
        )
        item.setData(Qt.CheckState.Unchecked, Qt.ItemDataRole.CheckStateRole)
        if icon_path:
            item.setData(
                QIcon(CustomIconEngine(icon_path, icon_color or "foreground")),
                Qt.ItemDataRole.DecorationRole,
            )
        cast(QStandardItemModel, self.model()).appendRow(item)
        self.update_text()

    @property
    def item_ids(self) -> list[ItemData]:
        """Return the list of ids checked by the user."""
        return [
            cast(QStandardItemModel, self.model()).item(i).data()
            for i in range(self.model().rowCount())
            if cast(QStandardItemModel, self.model()).item(i).checkState()
            == Qt.CheckState.Checked
        ]

    @item_ids.setter
    def item_ids(self, item_ids: list[ItemData]) -> None:
        """Set the ids that are selected/checked."""
        if self.single_mode and len(item_ids) > 1:
            raise TooManyItemsError(
                "Passed list contains more than one "
                "item for single mode checkbox."
            )
        log.debug("Setting item_ids: %s", item_ids)
        for idx in range(self.model().rowCount()):
            data = cast(QStandardItemModel, self.model()).item(idx).data()
            cast(QStandardItemModel, self.model()).item(idx).setCheckState(
                Qt.CheckState.Checked
                if data in item_ids
                else Qt.CheckState.Unchecked
            )


class SelectAllComboBox(CheckableComboBox[str | int], Generic[ItemData]):
    """CheckableComboBox that has an entry that alters the whole selection."""

    def __init__(self, parent: QWidget | None = None) -> None:
        """Create a new SelectAllComboBox."""
        super().__init__(parent)
        self.addItem(self.tr("Alles auswählen"))
        self.model().item(0).setAutoTristate(True)
        self.insertSeparator(1)

    def initialize_content(self) -> None:
        """Select or deselect all items based on "select all" item."""
        self.handle_data_change(
            self.model().index(0, 0, QModelIndex()),
            self.model().index(0, 0, QModelIndex()),
            [Qt.ItemDataRole.CheckStateRole],
        )

    @Slot(
        QModelIndex,
        QModelIndex,
        "QVector<int>",  # type: ignore
        name="handle_data_change",
    )
    def handle_data_change(
        self,
        start: QModelIndex,
        end: QModelIndex,
        roles: list[Qt.ItemDataRole],
    ) -> None:
        """Handle click on "Select all"."""
        select_all_item = self.model().item(0)
        self.model().blockSignals(True)
        # handle how the select_all_item affects all other items
        if start == self.model().index(0, 0, QModelIndex()):
            if select_all_item.checkState() == Qt.CheckState.Checked:
                log.debug("Selecting all items.")
                for idx in range(2, self.model().rowCount()):
                    self.model().item(idx).setCheckState(Qt.CheckState.Checked)
            elif select_all_item.checkState() == Qt.CheckState.Unchecked:
                log.debug("Deselecting all items.")
                for idx in range(2, self.model().rowCount()):
                    self.model().item(idx).setCheckState(
                        Qt.CheckState.Unchecked
                    )
            elif (
                select_all_item.checkState() == Qt.CheckState.PartiallyChecked
            ):
                return
        # handle how all other items affect the select_all_item
        elif all(
            self.model().item(idx).checkState() == Qt.CheckState.Checked
            for idx in range(2, self.model().rowCount())
        ):
            select_all_item.setCheckState(Qt.CheckState.Checked)
        elif all(
            self.model().item(idx).checkState() == Qt.CheckState.Unchecked
            for idx in range(2, self.model().rowCount())
        ):
            select_all_item.setCheckState(Qt.CheckState.Unchecked)
        else:
            select_all_item.setCheckState(Qt.CheckState.PartiallyChecked)
        self.model().blockSignals(False)
        super().handle_data_change(
            self.model().index(2, 0, QModelIndex()), end, roles
        )

    def _get_text(self) -> str:
        """Return the text that is shown at the top of the combobox."""
        texts = [
            str(cast(QStandardItemModel, self.model()).item(idx).data())
            for idx in range(2, self.model().rowCount())
            if cast(QStandardItemModel, self.model()).item(idx).checkState()
            == Qt.CheckState.Checked
        ]

        text = ", ".join(texts)
        # set to no index otherwise the state icon is displayed from
        # the first item in the combo box text which is selected by default
        self.setCurrentIndex(-1)
        if not text:
            text = self._default_text
        return text



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'test_window.ui'
##
## Created by: Qt User Interface Compiler version 6.4.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QCheckBox,
    QDateTimeEdit,
    QDial,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QProgressBar,
    QPushButton,
    QRadioButton,
    QScrollBar,
    QSizePolicy,
    QSlider,
    QSpacerItem,
    QSpinBox,
    QSplitter,
    QTableWidget,
    QTableWidgetItem,
    QTabWidget,
    QTextEdit,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)

from qute_style.widgets.icon_button import IconButton
from qute_style.widgets.styled_combobox import StyledComboBox
from qute_style.widgets.toggle import Toggle
from qute_style_examples.sample_classes import (
    SelectAllTestComboBox,
    TestComboBox,
)


class Ui_test_widget(object):
    def setupUi(self, test_widget):
        if not test_widget.objectName():
            test_widget.setObjectName("test_widget")
        test_widget.resize(1219, 886)
        self.gridLayout_2 = QGridLayout(test_widget)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.groupBox_3 = QGroupBox(test_widget)
        self.groupBox_3.setObjectName("groupBox_3")
        self.groupBox_3.setCheckable(False)
        self.gridLayout = QGridLayout(self.groupBox_3)
        self.gridLayout.setObjectName("gridLayout")
        self.lineEdit = QLineEdit(self.groupBox_3)
        self.lineEdit.setObjectName("lineEdit")
        self.lineEdit.setEchoMode(QLineEdit.Normal)

        self.gridLayout.addWidget(self.lineEdit, 6, 0, 1, 2)

        self.horizontalScrollBar = QScrollBar(self.groupBox_3)
        self.horizontalScrollBar.setObjectName("horizontalScrollBar")
        self.horizontalScrollBar.setValue(25)
        self.horizontalScrollBar.setOrientation(Qt.Horizontal)

        self.gridLayout.addWidget(self.horizontalScrollBar, 10, 1, 1, 1)

        self.label_3 = QLabel(self.groupBox_3)
        self.label_3.setObjectName("label_3")

        self.gridLayout.addWidget(self.label_3, 11, 0, 1, 1)

        self.dateTimeEdit = QDateTimeEdit(self.groupBox_3)
        self.dateTimeEdit.setObjectName("dateTimeEdit")

        self.gridLayout.addWidget(self.dateTimeEdit, 8, 0, 1, 2)

        self.label_2 = QLabel(self.groupBox_3)
        self.label_2.setObjectName("label_2")

        self.gridLayout.addWidget(self.label_2, 3, 0, 1, 1)

        self.dial = QDial(self.groupBox_3)
        self.dial.setObjectName("dial")
        self.dial.setValue(30)

        self.gridLayout.addWidget(self.dial, 10, 0, 1, 1)

        self.styled_combobox = StyledComboBox(self.groupBox_3)
        self.styled_combobox.addItem("")
        self.styled_combobox.addItem("")
        self.styled_combobox.addItem("")
        self.styled_combobox.setObjectName("styled_combobox")

        self.gridLayout.addWidget(self.styled_combobox, 0, 1, 1, 1)

        self.transparent_combobox = StyledComboBox(self.groupBox_3)
        self.transparent_combobox.addItem("")
        self.transparent_combobox.addItem("")
        self.transparent_combobox.addItem("")
        self.transparent_combobox.setObjectName("transparent_combobox")

        self.gridLayout.addWidget(self.transparent_combobox, 3, 1, 1, 1)

        self.label_4 = QLabel(self.groupBox_3)
        self.label_4.setObjectName("label_4")

        self.gridLayout.addWidget(self.label_4, 1, 0, 1, 1)

        self.checkable_combobox = TestComboBox(self.groupBox_3)
        self.checkable_combobox.setObjectName("checkable_combobox")

        self.gridLayout.addWidget(self.checkable_combobox, 1, 1, 1, 1)

        self.horizontalSlider = QSlider(self.groupBox_3)
        self.horizontalSlider.setObjectName("horizontalSlider")
        self.horizontalSlider.setMaximum(100)
        self.horizontalSlider.setValue(30)
        self.horizontalSlider.setOrientation(Qt.Horizontal)

        self.gridLayout.addWidget(self.horizontalSlider, 11, 1, 1, 1)

        self.spinBox = QSpinBox(self.groupBox_3)
        self.spinBox.setObjectName("spinBox")

        self.gridLayout.addWidget(self.spinBox, 7, 0, 1, 2)

        self.label = QLabel(self.groupBox_3)
        self.label.setObjectName("label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.label_5 = QLabel(self.groupBox_3)
        self.label_5.setObjectName("label_5")

        self.gridLayout.addWidget(self.label_5, 2, 0, 1, 1)

        self.comboBox = SelectAllTestComboBox(self.groupBox_3)
        self.comboBox.setObjectName("comboBox")

        self.gridLayout.addWidget(self.comboBox, 2, 1, 1, 1)

        self.gridLayout_2.addWidget(self.groupBox_3, 4, 1, 1, 1)

        self.groupBox_2 = QGroupBox(test_widget)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_3 = QVBoxLayout(self.groupBox_2)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.pushButton = QPushButton(self.groupBox_2)
        self.pushButton.setObjectName("pushButton")

        self.verticalLayout_3.addWidget(self.pushButton)

        self.pushButton_2 = QPushButton(self.groupBox_2)
        self.pushButton_2.setObjectName("pushButton_2")
        self.pushButton_2.setCheckable(True)

        self.verticalLayout_3.addWidget(self.pushButton_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.pushButton_3 = IconButton(self.groupBox_2)
        self.pushButton_3.setObjectName("pushButton_3")
        self.pushButton_3.setFlat(False)

        self.horizontalLayout_3.addWidget(self.pushButton_3)

        self.horizontalSpacer_2 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_3.addItem(self.horizontalSpacer_2)

        self.verticalLayout_3.addLayout(self.horizontalLayout_3)

        self.pushButton_4 = IconButton(self.groupBox_2)
        self.pushButton_4.setObjectName("pushButton_4")

        self.verticalLayout_3.addWidget(self.pushButton_4)

        self.gridLayout_2.addWidget(self.groupBox_2, 2, 1, 1, 1)

        self.groupBox = QGroupBox(test_widget)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_4 = QVBoxLayout(self.groupBox)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout_5 = QVBoxLayout()
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.radioButton = QRadioButton(self.groupBox)
        self.radioButton.setObjectName("radioButton")
        self.radioButton.setToolTipDuration(-10)
        self.radioButton.setChecked(True)

        self.verticalLayout_5.addWidget(self.radioButton)

        self.radioButton_2 = QRadioButton(self.groupBox)
        self.radioButton_2.setObjectName("radioButton_2")

        self.verticalLayout_5.addWidget(self.radioButton_2)

        self.radioButton_3 = QRadioButton(self.groupBox)
        self.radioButton_3.setObjectName("radioButton_3")

        self.verticalLayout_5.addWidget(self.radioButton_3)

        self.checkBox_5 = QCheckBox(self.groupBox)
        self.checkBox_5.setObjectName("checkBox_5")
        self.checkBox_5.setTristate(True)

        self.verticalLayout_5.addWidget(self.checkBox_5)

        self.custom_icon_engine_checkbox = QCheckBox(self.groupBox)
        self.custom_icon_engine_checkbox.setObjectName(
            "custom_icon_engine_checkbox"
        )

        self.verticalLayout_5.addWidget(self.custom_icon_engine_checkbox)

        self.icon_checkbox = QCheckBox(self.groupBox)
        self.icon_checkbox.setObjectName("icon_checkbox")

        self.verticalLayout_5.addWidget(self.icon_checkbox)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.checkBox_3 = Toggle(self.groupBox)
        self.checkBox_3.setObjectName("checkBox_3")

        self.horizontalLayout_5.addWidget(self.checkBox_3)

        self.horizontalSpacer_3 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_5.addItem(self.horizontalSpacer_3)

        self.checkBox_4 = Toggle(self.groupBox)
        self.checkBox_4.setObjectName("checkBox_4")
        self.checkBox_4.setLayoutDirection(Qt.RightToLeft)

        self.horizontalLayout_5.addWidget(self.checkBox_4)

        self.verticalLayout_5.addLayout(self.horizontalLayout_5)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.checkBox = Toggle(self.groupBox)
        self.checkBox.setObjectName("checkBox")
        self.checkBox.setCheckable(True)
        self.checkBox.setChecked(False)
        self.checkBox.setAutoExclusive(False)

        self.horizontalLayout_4.addWidget(self.checkBox)

        self.horizontalSpacer_4 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_4.addItem(self.horizontalSpacer_4)

        self.checkBox_2 = Toggle(self.groupBox)
        self.checkBox_2.setObjectName("checkBox_2")
        self.checkBox_2.setLayoutDirection(Qt.RightToLeft)

        self.horizontalLayout_4.addWidget(self.checkBox_2)

        self.verticalLayout_5.addLayout(self.horizontalLayout_4)

        self.verticalLayout_4.addLayout(self.verticalLayout_5)

        self.gridLayout_2.addWidget(self.groupBox, 2, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.disable_widgets = QCheckBox(test_widget)
        self.disable_widgets.setObjectName("disable_widgets")
        font = QFont()
        font.setFamilies(["Segoe UI"])
        self.disable_widgets.setFont(font)
        self.disable_widgets.setChecked(False)

        self.horizontalLayout.addWidget(self.disable_widgets)

        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.gridLayout_2.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.progressBar = QProgressBar(test_widget)
        self.progressBar.setObjectName("progressBar")
        self.progressBar.setValue(30)

        self.gridLayout_2.addWidget(self.progressBar, 6, 0, 1, 2)

        self.tabWidget = QTabWidget(test_widget)
        self.tabWidget.setObjectName("tabWidget")
        self.tabWidget.setAcceptDrops(True)
        self.tabWidget.setAutoFillBackground(False)
        self.tab = QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout = QVBoxLayout(self.tab)
        self.verticalLayout.setObjectName("verticalLayout")
        self.textEdit = QTextEdit(self.tab)
        self.textEdit.setObjectName("textEdit")

        self.verticalLayout.addWidget(self.textEdit)

        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName("tab_2")
        self.verticalLayout_2 = QVBoxLayout(self.tab_2)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tableWidget = QTableWidget(self.tab_2)
        if self.tableWidget.columnCount() < 10:
            self.tableWidget.setColumnCount(10)
        __qtablewidgetitem = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        __qtablewidgetitem3 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(3, __qtablewidgetitem3)
        __qtablewidgetitem4 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(4, __qtablewidgetitem4)
        __qtablewidgetitem5 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(5, __qtablewidgetitem5)
        __qtablewidgetitem6 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(6, __qtablewidgetitem6)
        __qtablewidgetitem7 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(7, __qtablewidgetitem7)
        __qtablewidgetitem8 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(8, __qtablewidgetitem8)
        __qtablewidgetitem9 = QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(9, __qtablewidgetitem9)
        if self.tableWidget.rowCount() < 10:
            self.tableWidget.setRowCount(10)
        __qtablewidgetitem10 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(0, __qtablewidgetitem10)
        __qtablewidgetitem11 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(1, __qtablewidgetitem11)
        __qtablewidgetitem12 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(2, __qtablewidgetitem12)
        __qtablewidgetitem13 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(3, __qtablewidgetitem13)
        __qtablewidgetitem14 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(4, __qtablewidgetitem14)
        __qtablewidgetitem15 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(5, __qtablewidgetitem15)
        __qtablewidgetitem16 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(6, __qtablewidgetitem16)
        __qtablewidgetitem17 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(7, __qtablewidgetitem17)
        __qtablewidgetitem18 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(8, __qtablewidgetitem18)
        __qtablewidgetitem19 = QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(9, __qtablewidgetitem19)
        __qtablewidgetitem20 = QTableWidgetItem()
        self.tableWidget.setItem(0, 0, __qtablewidgetitem20)
        __qtablewidgetitem21 = QTableWidgetItem()
        self.tableWidget.setItem(0, 1, __qtablewidgetitem21)
        __qtablewidgetitem22 = QTableWidgetItem()
        self.tableWidget.setItem(1, 0, __qtablewidgetitem22)
        self.tableWidget.setObjectName("tableWidget")

        self.verticalLayout_2.addWidget(self.tableWidget)

        self.tabWidget.addTab(self.tab_2, "")
        self.tab_3 = QWidget()
        self.tab_3.setObjectName("tab_3")
        self.tab_3.setCursor(QCursor(Qt.ArrowCursor))
        self.horizontalLayout_2 = QHBoxLayout(self.tab_3)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.splitter = QSplitter(self.tab_3)
        self.splitter.setObjectName("splitter")
        self.splitter.setOrientation(Qt.Horizontal)
        self.splitter_label = QLabel(self.splitter)
        self.splitter_label.setObjectName("splitter_label")
        self.splitter.addWidget(self.splitter_label)
        self.splitter_button = QPushButton(self.splitter)
        self.splitter_button.setObjectName("splitter_button")
        self.splitter.addWidget(self.splitter_button)

        self.horizontalLayout_2.addWidget(self.splitter)

        self.tabWidget.addTab(self.tab_3, "")
        self.tab_4 = QWidget()
        self.tab_4.setObjectName("tab_4")
        self.verticalLayout_6 = QVBoxLayout(self.tab_4)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.drop_widget = QTreeWidget(self.tab_4)
        self.drop_widget.setObjectName("drop_widget")
        self.drop_widget.setAcceptDrops(True)
        self.drop_widget.setDragEnabled(True)
        self.drop_widget.setDragDropMode(QAbstractItemView.DragDrop)

        self.verticalLayout_6.addWidget(self.drop_widget)

        self.clear_drop_button = QPushButton(self.tab_4)
        self.clear_drop_button.setObjectName("clear_drop_button")

        self.verticalLayout_6.addWidget(self.clear_drop_button)

        self.tabWidget.addTab(self.tab_4, "")
        self.tab_6 = QWidget()
        self.tab_6.setObjectName("tab_6")
        self.horizontalLayout_7 = QHBoxLayout(self.tab_6)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.verticalLayout_7 = QVBoxLayout()
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.treeWidget = QTreeWidget(self.tab_6)
        __qtreewidgetitem = QTreeWidgetItem(self.treeWidget)
        __qtreewidgetitem1 = QTreeWidgetItem(__qtreewidgetitem)
        QTreeWidgetItem(__qtreewidgetitem1)
        QTreeWidgetItem(self.treeWidget)
        self.treeWidget.setObjectName("treeWidget")

        self.verticalLayout_7.addWidget(self.treeWidget)

        self.horizontalLayout_7.addLayout(self.verticalLayout_7)

        self.tabWidget.addTab(self.tab_6, "")

        self.gridLayout_2.addWidget(self.tabWidget, 4, 0, 1, 1)

        self.retranslateUi(test_widget)

        self.pushButton_4.setDefault(False)
        self.tabWidget.setCurrentIndex(4)

    # setupUi

    def retranslateUi(self, test_widget):
        test_widget.setWindowTitle(
            QCoreApplication.translate("test_widget", "Form", None)
        )
        self.groupBox_3.setTitle(
            QCoreApplication.translate("test_widget", "GroupBox", None)
        )
        self.lineEdit.setText(
            QCoreApplication.translate("test_widget", "test", None)
        )
        self.label_3.setText(
            QCoreApplication.translate(
                "test_widget", "Test QProgressBar:", None
            )
        )
        self.label_2.setText(
            QCoreApplication.translate(
                "test_widget", "Transparent ComboBox", None
            )
        )
        self.styled_combobox.setItemText(
            0, QCoreApplication.translate("test_widget", "New Item", None)
        )
        self.styled_combobox.setItemText(
            1, QCoreApplication.translate("test_widget", "New Item 2", None)
        )
        self.styled_combobox.setItemText(
            2, QCoreApplication.translate("test_widget", "New Item 3", None)
        )

        self.transparent_combobox.setItemText(
            0, QCoreApplication.translate("test_widget", "New Item", None)
        )
        self.transparent_combobox.setItemText(
            1, QCoreApplication.translate("test_widget", "New Item 2", None)
        )
        self.transparent_combobox.setItemText(
            2, QCoreApplication.translate("test_widget", "New Item 3", None)
        )

        self.label_4.setText(
            QCoreApplication.translate(
                "test_widget", "Checkable ComboBox", None
            )
        )
        self.label.setText(
            QCoreApplication.translate("test_widget", "Styled ComboBox", None)
        )
        self.label_5.setText(
            QCoreApplication.translate(
                "test_widget", "SelectAll ComboBox", None
            )
        )
        self.groupBox_2.setTitle(
            QCoreApplication.translate("test_widget", "GroupBox", None)
        )
        self.pushButton.setText(
            QCoreApplication.translate("test_widget", "PushButton", None)
        )
        self.pushButton_2.setText(
            QCoreApplication.translate("test_widget", "PushButton", None)
        )
        self.pushButton_3.setText(
            QCoreApplication.translate("test_widget", "Test-Text", None)
        )
        self.pushButton_4.setText("")
        self.groupBox.setTitle(
            QCoreApplication.translate("test_widget", "GroupBox", None)
        )
        self.radioButton.setText(
            QCoreApplication.translate("test_widget", "RadioButton", None)
        )
        self.radioButton_2.setText(
            QCoreApplication.translate("test_widget", "RadioButton", None)
        )
        self.radioButton_3.setText(
            QCoreApplication.translate("test_widget", "RadioButton", None)
        )
        self.checkBox_5.setText(
            QCoreApplication.translate("test_widget", "CheckBox", None)
        )
        self.custom_icon_engine_checkbox.setText(
            QCoreApplication.translate(
                "test_widget", "Checkbox with CustomIconEngine", None
            )
        )
        self.icon_checkbox.setText(
            QCoreApplication.translate(
                "test_widget", "CheckBox with QIcon", None
            )
        )
        self.checkBox_3.setText(
            QCoreApplication.translate(
                "test_widget",
                "This is a very long test text so that we can crop",
                None,
            )
        )
        self.checkBox_4.setText(
            QCoreApplication.translate("test_widget", "CheckBox", None)
        )
        self.checkBox.setText("")
        self.checkBox_2.setText("")
        self.disable_widgets.setText(
            QCoreApplication.translate("test_widget", "Disable Widgets", None)
        )
        self.textEdit.setHtml(
            QCoreApplication.translate(
                "test_widget",
                '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">\n'
                '<html><head><meta name="qrichtext" content="1" /><style type="text/css">\n'
                "p, li { white-space: pre-wrap; }\n"
                "</style></head><body style=\" font-family:'MS Shell Dlg 2'; font-size:8.25pt; font-weight:400; font-style:normal;\">\n"
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;"><span style=" font-family:\'.SF NS Text\'; font-size:13pt;">test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />'
                "test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br /></span></p></body></html>",
                None,
            )
        )
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab),
            QCoreApplication.translate("test_widget", "TextEdit", None),
        )
        ___qtablewidgetitem = self.tableWidget.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(
            QCoreApplication.translate("test_widget", "1", None)
        )
        ___qtablewidgetitem1 = self.tableWidget.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(
            QCoreApplication.translate("test_widget", "2", None)
        )
        ___qtablewidgetitem2 = self.tableWidget.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(
            QCoreApplication.translate("test_widget", "3", None)
        )
        ___qtablewidgetitem3 = self.tableWidget.horizontalHeaderItem(3)
        ___qtablewidgetitem3.setText(
            QCoreApplication.translate("test_widget", "4", None)
        )
        ___qtablewidgetitem4 = self.tableWidget.horizontalHeaderItem(4)
        ___qtablewidgetitem4.setText(
            QCoreApplication.translate("test_widget", "5", None)
        )
        ___qtablewidgetitem5 = self.tableWidget.horizontalHeaderItem(5)
        ___qtablewidgetitem5.setText(
            QCoreApplication.translate("test_widget", "6", None)
        )
        ___qtablewidgetitem6 = self.tableWidget.horizontalHeaderItem(6)
        ___qtablewidgetitem6.setText(
            QCoreApplication.translate("test_widget", "7", None)
        )
        ___qtablewidgetitem7 = self.tableWidget.horizontalHeaderItem(7)
        ___qtablewidgetitem7.setText(
            QCoreApplication.translate("test_widget", "8", None)
        )
        ___qtablewidgetitem8 = self.tableWidget.horizontalHeaderItem(8)
        ___qtablewidgetitem8.setText(
            QCoreApplication.translate("test_widget", "9", None)
        )
        ___qtablewidgetitem9 = self.tableWidget.horizontalHeaderItem(9)
        ___qtablewidgetitem9.setText(
            QCoreApplication.translate("test_widget", "10", None)
        )
        ___qtablewidgetitem10 = self.tableWidget.verticalHeaderItem(0)
        ___qtablewidgetitem10.setText(
            QCoreApplication.translate("test_widget", "1", None)
        )
        ___qtablewidgetitem11 = self.tableWidget.verticalHeaderItem(1)
        ___qtablewidgetitem11.setText(
            QCoreApplication.translate("test_widget", "2", None)
        )
        ___qtablewidgetitem12 = self.tableWidget.verticalHeaderItem(2)
        ___qtablewidgetitem12.setText(
            QCoreApplication.translate("test_widget", "3", None)
        )
        ___qtablewidgetitem13 = self.tableWidget.verticalHeaderItem(3)
        ___qtablewidgetitem13.setText(
            QCoreApplication.translate("test_widget", "4", None)
        )
        ___qtablewidgetitem14 = self.tableWidget.verticalHeaderItem(4)
        ___qtablewidgetitem14.setText(
            QCoreApplication.translate("test_widget", "5", None)
        )
        ___qtablewidgetitem15 = self.tableWidget.verticalHeaderItem(5)
        ___qtablewidgetitem15.setText(
            QCoreApplication.translate("test_widget", "6", None)
        )
        ___qtablewidgetitem16 = self.tableWidget.verticalHeaderItem(6)
        ___qtablewidgetitem16.setText(
            QCoreApplication.translate("test_widget", "7", None)
        )
        ___qtablewidgetitem17 = self.tableWidget.verticalHeaderItem(7)
        ___qtablewidgetitem17.setText(
            QCoreApplication.translate("test_widget", "8", None)
        )
        ___qtablewidgetitem18 = self.tableWidget.verticalHeaderItem(8)
        ___qtablewidgetitem18.setText(
            QCoreApplication.translate("test_widget", "9", None)
        )
        ___qtablewidgetitem19 = self.tableWidget.verticalHeaderItem(9)
        ___qtablewidgetitem19.setText(
            QCoreApplication.translate("test_widget", "10", None)
        )

        __sortingEnabled = self.tableWidget.isSortingEnabled()
        self.tableWidget.setSortingEnabled(False)
        ___qtablewidgetitem20 = self.tableWidget.item(0, 0)
        ___qtablewidgetitem20.setText(
            QCoreApplication.translate("test_widget", "Milben Power", None)
        )
        ___qtablewidgetitem21 = self.tableWidget.item(0, 1)
        ___qtablewidgetitem21.setText(
            QCoreApplication.translate("test_widget", "Milben Power", None)
        )
        ___qtablewidgetitem22 = self.tableWidget.item(1, 0)
        ___qtablewidgetitem22.setText(
            QCoreApplication.translate("test_widget", "Miben Power", None)
        )
        self.tableWidget.setSortingEnabled(__sortingEnabled)

        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_2),
            QCoreApplication.translate("test_widget", "Tab Table", None),
        )
        self.splitter_label.setText(
            QCoreApplication.translate("test_widget", "TextLabel", None)
        )
        self.splitter_button.setText(
            QCoreApplication.translate("test_widget", "PushButton", None)
        )
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_3),
            QCoreApplication.translate("test_widget", "QSplitter", None),
        )
        ___qtreewidgetitem = self.drop_widget.headerItem()
        ___qtreewidgetitem.setText(
            0, QCoreApplication.translate("test_widget", "Test Files", None)
        )
        self.clear_drop_button.setText(
            QCoreApplication.translate("test_widget", "Clear", None)
        )
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_4),
            QCoreApplication.translate("test_widget", "DropArea", None),
        )
        ___qtreewidgetitem1 = self.treeWidget.headerItem()
        ___qtreewidgetitem1.setText(
            0, QCoreApplication.translate("test_widget", "1", None)
        )

        __sortingEnabled1 = self.treeWidget.isSortingEnabled()
        self.treeWidget.setSortingEnabled(False)
        ___qtreewidgetitem2 = self.treeWidget.topLevelItem(0)
        ___qtreewidgetitem2.setText(
            0, QCoreApplication.translate("test_widget", "Item 1", None)
        )
        ___qtreewidgetitem3 = ___qtreewidgetitem2.child(0)
        ___qtreewidgetitem3.setText(
            0, QCoreApplication.translate("test_widget", "Subitem", None)
        )
        ___qtreewidgetitem4 = ___qtreewidgetitem3.child(0)
        ___qtreewidgetitem4.setText(
            0, QCoreApplication.translate("test_widget", "SubSub", None)
        )
        ___qtreewidgetitem5 = self.treeWidget.topLevelItem(1)
        ___qtreewidgetitem5.setText(
            0, QCoreApplication.translate("test_widget", "Item 2", None)
        )
        self.treeWidget.setSortingEnabled(__sortingEnabled1)

        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_6),
            QCoreApplication.translate("test_widget", "TreeView", None),
        )

    # retranslateUi



"""QuteStyle Library - helper: useful functions for QuteStyle tools."""
import logging
import sys
from pathlib import Path
from typing import Union
from xml.etree.ElementTree import Element, SubElement, tostring

from PySide6.QtGui import QColor, QPaintEvent
from PySide6.QtWidgets import QWidget

from qute_style.style import get_color
from qute_style.widgets.spinner import WaitingSpinner

log = logging.getLogger("qute_style")


def check_ide() -> bool:
    """Check if the application is run from the IDE."""
    return Path(sys.argv[0]).suffix == ".py"


class StyledWaitingSpinner(WaitingSpinner):
    """Styled Version of QWaitingSpinner."""

    def paintEvent(self, _: QPaintEvent) -> None:
        """Overwrite method to change color of spinner."""
        self._color = QColor(get_color("context_color"))
        super().paintEvent(_)


def create_waiting_spinner(
    parent: QWidget,
    number_of_lines: int = 28,
    line_length: int = 20,
    inner_radius: int = 15,
) -> StyledWaitingSpinner:
    """Create a waiting spinner with default config."""
    spinner = StyledWaitingSpinner(parent)
    spinner.color = QColor(get_color("context_color"))
    spinner.number_of_lines = number_of_lines
    spinner.line_length = line_length
    spinner.inner_radius = inner_radius
    spinner.line_width = 2
    return spinner


def create_tooltip(title: str, description: Union[str, list[str]]) -> str:
    """Create a tooltip as HTML str."""
    top = Element("div")
    first = SubElement(top, "p")
    headline = SubElement(first, "b")
    headline.text = title
    second = SubElement(top, "p")
    text = SubElement(second, "small")
    if isinstance(description, str):
        text.text = description
    else:
        text.text = "<br />".join(description)
    return tostring(top).decode("utf-8")



"""QuteStyleApplication."""
from __future__ import annotations

import logging
import operator
from copy import copy
from typing import Type

from PySide6 import QtCore
from PySide6.QtCore import QRectF, QSize, Qt, Slot
from PySide6.QtGui import (
    QCloseEvent,
    QColor,
    QMouseEvent,
    QPainter,
    QPainterPath,
    QPixmap,
)
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (
    QApplication,
    QGridLayout,
    QLabel,
    QMainWindow,
    QSplashScreen,
)

from qute_style.helper import check_ide, create_waiting_spinner
from qute_style.qs_main_window import AppData, CustomMainWindow
from qute_style.startup_threads import StartupThread
from qute_style.style import get_color, get_style

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name


class CustomSplashScreen(QSplashScreen):
    """Custom Splash screen for startup of app."""

    def __init__(self, app_data: AppData) -> None:
        """Init CustomSplashScreen."""
        super().__init__()
        style_sheet = get_style()
        self.setStyleSheet(style_sheet)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)
        background_color = get_color("bg_one")
        # use a pixmap as background
        size = QSize(350, 200)
        self.setFixedSize(size)
        pixmap = QPixmap(size)
        pixmap.fill(Qt.GlobalColor.transparent)
        painter = QPainter(pixmap)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        path = QPainterPath()
        path.addRoundedRect(QRectF(0, 0, size.width(), size.height()), 12, 12)
        painter.fillPath(path, QColor(background_color))
        painter.end()
        # working solution to get the rounded corners to a splash screen
        self.setPixmap(pixmap)
        grid_layout = QGridLayout(self)
        grid_layout.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        logo_svg = QSvgWidget(app_data.app_splash_icon, self)
        logo_svg.setFixedSize(QSize(140, 140))
        grid_layout.addWidget(
            logo_svg, 0, 0, 1, 0, alignment=Qt.AlignmentFlag.AlignCenter
        )
        grid_layout.addWidget(
            QLabel(f"Starting {app_data.app_name}. Please wait", self), 1, 0
        )
        label = QLabel(self)
        label.setFixedSize(30, 30)
        self._spinner = create_waiting_spinner(label, 18, 10, 5)
        self._spinner.start()
        grid_layout.addWidget(label, 1, 1)

    def closeEvent(  # pylint: disable=invalid-name
        self, close_event: QCloseEvent
    ) -> None:
        """Handle close event."""
        self._spinner.stop()
        super().closeEvent(close_event)

    def mousePressEvent(  # pylint: disable=invalid-name
        self, event: QMouseEvent
    ) -> None:
        """Disable default "click-to-dismiss" behaviour."""


class QuteStyleApplication(  # pylint: disable=too-many-instance-attributes
    QApplication
):
    """QuteStyleApplication."""

    MAIN_WINDOW_CLASS: Type[CustomMainWindow]

    STARTUP_THREADS: list[Type[StartupThread]] = []

    APP_DATA: AppData

    def __init__(self, argv: list[str], show_splash: bool = True) -> None:
        """Init QuteStyleApplication."""
        super().__init__(argv)

        self._update = "-u" not in argv and not check_ide()
        self._force_whats_new = "-w" in argv
        self._reset_settings = "-c" in argv

        self.check_startup_thread_configuration()
        self.setApplicationName(self.APP_DATA.app_name)
        self.setOrganizationName(self.APP_DATA.organization_name)
        self.setOrganizationDomain(self.APP_DATA.organization_domain)
        if show_splash:
            self._splash_screen: QSplashScreen | None = CustomSplashScreen(
                self.APP_DATA
            )
            self._splash_screen.show()
        else:
            self._splash_screen = None
        self._main_window: QMainWindow | None = None

        self._threads_to_run: list[Type[StartupThread]] = copy(
            self.STARTUP_THREADS
        )

        # need to hold a reference to the threads until finished
        self._threads_running: list[StartupThread] = []
        self._threads_finished: list[StartupThread] = []

        self._handle_startup_threads()

    def check_startup_thread_configuration(self) -> None:
        """
        Check the configuration of the startup threads.

        In case exit application function is overriden check for a unique
        priority of EXIT_FUNCTION_PRIORITY.
        """
        thread_priorities: list[int] = []
        for thread in self.STARTUP_THREADS:
            if id(StartupThread.exit_application) != id(
                thread.exit_application
            ):
                assert (
                    thread.EXIT_FUNCTION_PRIORITY > 0
                ), "EXIT_FUNCTION_PRIORITY of startup threads must be > 0"
                assert (
                    thread.EXIT_FUNCTION_PRIORITY not in thread_priorities
                ), "EXIT_FUNCTION_PRIORITY of startup threads must be unique"
                thread_priorities.append(thread.EXIT_FUNCTION_PRIORITY)

    def _handle_startup_threads(self) -> None:
        """Handle startup threads if any."""
        # reimplement this method to handle starting threads or for
        # example removing threads before start
        self.show_main_window()

    def _run_threads(self) -> None:
        """Start processing the startup threads."""
        for thread_class in copy(self._threads_to_run):
            log.debug("Preparing thread to start: %s", thread_class)

            threads_finished = [
                thread.__class__ for thread in self._threads_finished
            ]
            if any(
                dependency not in threads_finished
                for dependency in thread_class.START_DEPENDS_ON
            ):
                log.debug("Cannot run thread, dependencies not met")
                continue
            thread = thread_class(self.APP_DATA)
            self._threads_running.append(thread)
            self._threads_to_run.remove(thread_class)
            thread.finished.connect(
                lambda t=thread: self.on_finished_thread(t)
            )
            thread.start()
            log.debug("Thread starting: %s", thread)

    @Slot(StartupThread, name="on_finished_thread")
    def on_finished_thread(self, thread: StartupThread) -> None:
        """Handle a finished StartupThread."""
        log.debug("Thread finished: %s", thread)
        self._threads_running.remove(thread)
        self._threads_finished.append(thread)
        if thread.exit_application:
            # clear all threads which are not started yet in case a thread
            # with exit is ended
            self._threads_to_run.clear()
        if not self._threads_running and not self._threads_to_run:
            # all running threads have finished. Check if there are threads
            # that exits the app
            exit_threads = [
                thread
                for thread in self._threads_finished
                if thread.exit_application
            ]
            # sort by EXIT_FUNCTION_PRIORITY
            exit_threads.sort(
                key=operator.attrgetter("EXIT_FUNCTION_PRIORITY")
            )
            self._threads_finished.clear()
            if exit_threads:
                log.debug("Running threads post process, Exit application.")
                exit_threads[0].exit_function()
                self.quit()
            else:
                log.debug(
                    "There are no threads left to run, "
                    "starting QuteStyleMainWindow."
                )
                self.show_main_window()
        else:
            log.debug("Threads still to run: %s", self._threads_to_run)
            log.debug("Thread currently running: %s", self._threads_running)
            self._run_threads()

    def show_main_window(self) -> None:
        """Show the main window."""
        self._main_window = self.MAIN_WINDOW_CLASS(
            self.APP_DATA, self._force_whats_new, self._reset_settings
        )
        if self._splash_screen:
            self._splash_screen.finish(self._main_window)
        self._main_window.show()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'whats_new_window.ui'
##
## Created by: Qt User Interface Compiler version 6.4.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QMenuBar,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QSpacerItem,
    QStatusBar,
    QVBoxLayout,
    QWidget,
)


class Ui_whats_new_window(object):
    def setupUi(self, whats_new_window):
        if not whats_new_window.objectName():
            whats_new_window.setObjectName("whats_new_window")
        whats_new_window.resize(389, 341)
        self.centralwidget = QWidget(whats_new_window)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea = QScrollArea(self.centralwidget)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 369, 249))
        self.horizontalLayout_3 = QHBoxLayout(self.scrollAreaWidgetContents)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.text_label = QLabel(self.scrollAreaWidgetContents)
        self.text_label.setObjectName("text_label")
        self.text_label.setWordWrap(True)

        self.horizontalLayout_3.addWidget(self.text_label)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.verticalLayout.addWidget(self.scrollArea)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.previous_button = QPushButton(self.centralwidget)
        self.previous_button.setObjectName("previous_button")

        self.horizontalLayout.addWidget(self.previous_button)

        self.next_button = QPushButton(self.centralwidget)
        self.next_button.setObjectName("next_button")

        self.horizontalLayout.addWidget(self.next_button)

        self.close_button = QPushButton(self.centralwidget)
        self.close_button.setObjectName("close_button")

        self.horizontalLayout.addWidget(self.close_button)

        self.verticalLayout.addLayout(self.horizontalLayout)

        whats_new_window.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(whats_new_window)
        self.menubar.setObjectName("menubar")
        self.menubar.setGeometry(QRect(0, 0, 389, 21))
        whats_new_window.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(whats_new_window)
        self.statusbar.setObjectName("statusbar")
        whats_new_window.setStatusBar(self.statusbar)

        self.retranslateUi(whats_new_window)

    # setupUi

    def retranslateUi(self, whats_new_window):
        whats_new_window.setWindowTitle(
            QCoreApplication.translate(
                "whats_new_window", "StyledMainWindow", None
            )
        )
        self.text_label.setText("")
        # if QT_CONFIG(tooltip)
        self.previous_button.setToolTip(
            QCoreApplication.translate(
                "whats_new_window", "Den letzten Eintrag anzeigen.", None
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.previous_button.setText(
            QCoreApplication.translate("whats_new_window", "Zur\u00fcck", None)
        )
        # if QT_CONFIG(tooltip)
        self.next_button.setToolTip(
            QCoreApplication.translate(
                "whats_new_window", "Den n\u00e4chsten Eintrag anzeigen.", None
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.next_button.setText(
            QCoreApplication.translate("whats_new_window", "Weiter", None)
        )
        # if QT_CONFIG(tooltip)
        self.close_button.setToolTip(
            QCoreApplication.translate(
                "whats_new_window",
                'Das Fenster schlie\u00dfen. Sie k\u00f6nnen das Fenster erneut anzeigen, indem Sie im Men\u00fc "Hilfe" den Eintrag "Neuerung" anw\u00e4hlen.',
                None,
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.close_button.setText(
            QCoreApplication.translate(
                "whats_new_window", "Schlie\u00dfen", None
            )
        )

    # retranslateUi
